package de.fraunhofer.iem.dataprovider.kpi.service

import de.fraunhofer.iem.dataprovider.kpi.dto.KPICreateDto
import de.fraunhofer.iem.dataprovider.kpi.dto.KPITreeChildResponseDto
import de.fraunhofer.iem.dataprovider.kpi.dto.KPITreeResponseDto
import de.fraunhofer.iem.dataprovider.kpi.entity.KPIEntity
import de.fraunhofer.iem.dataprovider.kpi.entity.KPIHierarchyEdgeEntity
import de.fraunhofer.iem.dataprovider.kpi.metrics.repositoryDetails.dto.RepositoryDetailsKpisDto
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIHierarchyEdgeRepository
import de.fraunhofer.iem.dataprovider.kpi.repository.KPIRepository
import de.fraunhofer.iem.dataprovider.kpi.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.MaximumKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RatioKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryEntity
import de.fraunhofer.iem.dataprovider.tools.occmd.dto.OccmdKpisDto
import org.springframework.data.repository.findByIdOrNull
import org.springframework.stereotype.Service
import java.util.*

@Service
class KPIService(
    private val kpiRepository: KPIRepository,
    private val kpiHierarchyEdgeRepository: KPIHierarchyEdgeRepository
) {

    private val logger = getLogger(javaClass)

    fun removeKPIChildrenLowerThanSecondLevel(kpiTreeResponseDto: KPITreeResponseDto): KPITreeResponseDto {
        val children: List<KPITreeChildResponseDto> = kpiTreeResponseDto.children.map {
            KPITreeChildResponseDto(
                KPITreeResponseDto(
                    it.kpi.id,
                    it.kpi.value,
                    it.kpi.name,
                    it.kpi.description,
                    it.kpi.isRoot,
                    it.kpi.displayValue,
                    emptyList(),
                    it.kpi.repositoryId,
                    it.kpi.isEmpty,
                ),
                it.weight
            )
        }

        return KPITreeResponseDto(
            kpiTreeResponseDto.id,
            kpiTreeResponseDto.value,
            kpiTreeResponseDto.name,
            kpiTreeResponseDto.description,
            kpiTreeResponseDto.isRoot,
            kpiTreeResponseDto.displayValue,
            children,
            kpiTreeResponseDto.repositoryId,
            kpiTreeResponseDto.isEmpty
        )
    }

    fun calculateAndSaveKpiTree(
        repository: RepositoryEntity,
        repositoryDetailsKpi: RepositoryDetailsKpisDto,
        vulnerabilityKpis: List<KPICreateDto>,
        occmdKpis: OccmdKpisDto
    ) {
        val rootKPI = generateKPITree(
            repositoryDetailsKpi = repositoryDetailsKpi,
            vulnerabilityKpis = vulnerabilityKpis,
            occmdKpis = occmdKpis
        )
        calculateKPIs(rootKPI)
        storeAndPurgeOld(repository, rootKPI)
    }

    private fun generateKPITree(
        repositoryDetailsKpi: RepositoryDetailsKpisDto,
        vulnerabilityKpis: List<KPICreateDto>,
        occmdKpis: OccmdKpisDto
    ): KPICreateDto {


        val signedCommitsRatioKPI =
            KPICreateDto(
                "Commit Signature Ratio",
                "Used to assess compliance with a common and transparent development process. It is desirable that all commits are signed by their authors. Therefore, the ratio of signed commits to all commits is determined to calculate this metric.",
                RatioKPICalculationStrategy()
            )
        if (repositoryDetailsKpi.numberOfCommitsKpi != null && repositoryDetailsKpi.numberOfSignedCommitsKPI != null) {
            signedCommitsRatioKPI.addChildKPI(repositoryDetailsKpi.numberOfCommitsKpi, 1.0)
            signedCommitsRatioKPI.addChildKPI(repositoryDetailsKpi.numberOfSignedCommitsKPI, 1.0)
        }

        val processComplianceKPI = KPICreateDto(
            "Process Compliance Score",
            "Assesses the development process of the software provided. For this purpose, the development process traceable in the repository is compared with common development standards to enable an assessment.",
            AggregationKPICalculationStrategy()
        )

        if (repositoryDetailsKpi.isDefaultBranchProtectedKPI != null) {
            processComplianceKPI.addChildKPI(repositoryDetailsKpi.isDefaultBranchProtectedKPI, 0.5)
        }
        processComplianceKPI.addChildKPI(signedCommitsRatioKPI, 0.5)

        val processTransparencyKPI = KPICreateDto(
            "Process Transparency Score",
            "Assesses the transparency resp. traceability of the development process of the provided software for external parties. For this purpose, various analyzes are performed that assess the availability of information about the software development process within the repository.",
            AggregationKPICalculationStrategy()
        )


        processTransparencyKPI.addChildKPI(signedCommitsRatioKPI, 1.0)


        val securityKPI = KPICreateDto(
            "Security Score",
            "Assesses the security of the software provided. For this purpose, various security-relevant analyzes are carried out, which, among other things, check the external dependencies or the code for vulnerabilities.",
            AggregationKPICalculationStrategy()
        )

        val maximalDependencyVulnerabilityKPI = KPICreateDto(
            "Maximal Dependency Vulnerability Score",
            "This score is calculated by the following formula: 100 - (max(CVSS score) * 10). Thus, a lower value indicates a more critical vulnerability.",
            // This strategy is tied to this specific KPI. It calculates 100 - max(children). Children are already multiplied by 10
            MaximumKPICalculationStrategy()
        )

        vulnerabilityKpis.forEach { maximalDependencyVulnerabilityKPI.addChildKPI(it, 0.0) }


        var dependencyWeight = 1.0
        if (occmdKpis.secretsKpi != null) {
            dependencyWeight -= 0.3
            securityKPI.addChildKPI(occmdKpis.secretsKpi, 0.3)
        }
        if (occmdKpis.checkedInBinaries != null) {
            dependencyWeight -= 0.2
            securityKPI.addChildKPI(occmdKpis.checkedInBinaries, 0.2)
        }
        securityKPI.addChildKPI(maximalDependencyVulnerabilityKPI, dependencyWeight)

        val rootKPI = KPICreateDto(
            "Project Score",
            "Assesses the project resp. the provided software in the aspects of maturity (based on quality, security and usability aspects) as well as development process.",
            AggregationKPICalculationStrategy(),
            isRoot = true,
        )
        rootKPI.addChildKPI(processTransparencyKPI, 0.25)
        rootKPI.addChildKPI(processComplianceKPI, 0.25)
        rootKPI.addChildKPI(securityKPI, 0.5)
        return rootKPI
    }


    private fun calculateKPIsRecursively(kpi: KPICreateDto, visited: MutableSet<KPICreateDto>) {
        // Check if the KPI has already been visited
        if (visited.contains(kpi)) {
            return
        }

        // Check if the KPI has child KPIs
        if (kpi.hierarchyEdges.isEmpty()) {
            // Leaf node, calculate the KPI value
            kpi.calculateKPI()
            visited.add(kpi)
            return
        }

        // Recursively calculate child KPIs first
        for (childEdge in kpi.hierarchyEdges) {
            calculateKPIsRecursively(childEdge.to, visited)
        }

        // Calculate the KPI value after processing child KPIs
        kpi.calculateKPI()
        visited.add(kpi)
    }

    fun calculateKPIs(rootKPI: KPICreateDto) {
        val visited: MutableSet<KPICreateDto> = mutableSetOf()
        calculateKPIsRecursively(rootKPI, visited)
    }


    fun findKPIById(id: UUID): KPIEntity? {
        return kpiRepository.findByIdOrNull(id)
    }

    fun findRootKPIByRepository(repositoryEntity: RepositoryEntity): KPIEntity? {
        val kpi = kpiRepository.findFirstByRepository_IdAndIsRootTrue(repositoryEntity.id!!)
        if (kpi.isEmpty) {
            return null
        }
        return kpi.get()
    }

    fun storeAndPurgeOld(repositoryEntity: RepositoryEntity, rootKPI: KPICreateDto) {
        purgeAllKPIs(repositoryEntity)
        val kpiEntityDtoToEntityMapping: MutableMap<KPICreateDto, KPIEntity> = mutableMapOf()
        storeKPI(repositoryEntity, rootKPI, kpiEntityDtoToEntityMapping)
    }

    private fun purgeAllKPIs(repositoryEntity: RepositoryEntity) {
        val kpiEntities: List<KPIEntity> = kpiRepository.findByRepository_Id(repositoryEntity.id!!)
        kpiRepository.deleteAll(kpiEntities)
    }

    private fun storeKPI(
        repositoryEntity: RepositoryEntity,
        kpi: KPICreateDto,
        kpiEntityDtoToEntityMapping: MutableMap<KPICreateDto, KPIEntity>
    ) {
        // already visited
        if (kpiEntityDtoToEntityMapping.contains(kpi)) {
            return
        }

        if (kpi.hierarchyEdges.isEmpty()) {
            // leaf node
            val kpiEntity = kpi.toDbObject()
            kpiEntity.repository = repositoryEntity

            kpiEntityDtoToEntityMapping[kpi] = kpiEntity
            kpiRepository.save(kpiEntity)
            logger.info("Storing leaf node ${kpi.name}")
            return
        }

        for (childEdge in kpi.hierarchyEdges) {
            storeKPI(repositoryEntity, childEdge.to, kpiEntityDtoToEntityMapping)
        }

        val kpiEntity = kpi.toDbObject()
        kpiEntity.repository = repositoryEntity
        kpiEntityDtoToEntityMapping[kpi] = kpiEntity
        kpiRepository.save(kpiEntity)
        logger.info("Storing node ${kpi.name}")
        for (hierarchyEdge in kpi.hierarchyEdges) {
            val hierarchyEdgeEntity = KPIHierarchyEdgeEntity()
            hierarchyEdgeEntity.from = kpiEntity
            hierarchyEdgeEntity.to = kpiEntityDtoToEntityMapping[hierarchyEdge.to]
            hierarchyEdgeEntity.weight = hierarchyEdge.weight
            kpiHierarchyEdgeRepository.save(hierarchyEdgeEntity)
            logger.info("Storing edge ${hierarchyEdge.weight}: From ${hierarchyEdge.from.name} to ${hierarchyEdge.to.name}")
        }
    }

}
