package de.fraunhofer.iem.dataprovider.repository.controller

import de.fraunhofer.iem.dataprovider.configuration.ApiPaths
import de.fraunhofer.iem.dataprovider.kpi.dto.KPITreeResponseDto
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.dto.RepositoryResponseDto
import de.fraunhofer.iem.dataprovider.repository.dto.ScoreCardResponseDto
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryEntity
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import org.springframework.web.bind.annotation.GetMapping
import org.springframework.web.bind.annotation.PathVariable
import org.springframework.web.bind.annotation.RestController
import java.util.*

@RestController
class RepositoryController(private val repositoryService: RepositoryService, private val kpiService: KPIService) {

    private val logger = getLogger(javaClass)

    @GetMapping(ApiPaths.REPOSITORY)
    suspend fun getAllRepositories(): List<RepositoryResponseDto> {
        logger.info("Get all repositories")
        return repositoryService.getAllRepositories().map { repositoryEntity: RepositoryEntity ->
            RepositoryResponseDto(
                repositoryEntity.id!!,
                repositoryEntity.name!!,
                repositoryEntity.repoId!!
            )
        }
    }

    @GetMapping(ApiPaths.REPOSITORY_SCORE_CARD)
    suspend fun getAllScoreCards(): List<ScoreCardResponseDto> {
        logger.info("Get all repositories as score cards")
        return repositoryService.getAllRepositories().map { repositoryEntity: RepositoryEntity ->
            var rootKPI = kpiService.findRootKPIByRepository(repositoryEntity)?.toResultDto()
            if (rootKPI != null) {
                rootKPI = kpiService.removeKPIChildrenLowerThanSecondLevel(rootKPI)
            }
            ScoreCardResponseDto(
                RepositoryResponseDto(
                    repositoryEntity.id!!,
                    repositoryEntity.name!!,
                    repositoryEntity.repoId!!
                ),
                rootKPI
            )
        }
    }

    @GetMapping(ApiPaths.SCORE_CARD_BY_REPOSITORY_ID)
    suspend fun getScoreCardByRepositoryId(@PathVariable id: UUID): ScoreCardResponseDto {
        logger.info("Get repository score card with id $id")
        val repositoryEntity = this.repositoryService.findRepoByID(id) ?: throw Exception("Repository can not be found")
        var rootKPI = kpiService.findRootKPIByRepository(repositoryEntity)?.toResultDto()
        if (rootKPI != null) {
            rootKPI = kpiService.removeKPIChildrenLowerThanSecondLevel(rootKPI)
        }
        return ScoreCardResponseDto(
            RepositoryResponseDto(
                repositoryEntity.id!!,
                repositoryEntity.name!!,
                repositoryEntity.repoId!!
            ),
            rootKPI
        )
    }

    @GetMapping(ApiPaths.REPOSITORY_ID)
    suspend fun getRepositoryById(@PathVariable id: UUID): RepositoryResponseDto {
        logger.info("Get repository with id $id")
        val repositoryEntity = this.repositoryService.findRepoByID(id) ?: throw Exception("Repository can not be found")
        return RepositoryResponseDto(
            repositoryEntity.id!!,
            repositoryEntity.name!!,
            repositoryEntity.repoId!!
        )
    }

    @GetMapping(ApiPaths.KPI_BY_REPOSITORY_ID)
    suspend fun getKPIByRepositoryId(@PathVariable id: UUID): KPITreeResponseDto {
        logger.info("Get KPI tree for repository id $id")
        val repositoryEntity = this.repositoryService.findRepoByID(id) ?: throw Exception("Repository can not be found")
        val rootKPI = this.kpiService.findRootKPIByRepository(repositoryEntity)
            ?: throw Exception("KPI tree has not been calculated yet")
        return rootKPI.toResultDto()
    }
}
